from datetime import datetime

__all__ = ['get_logs']

def get_logs(driver, log_type):
    raw_logs = driver.get_log(log_type)

    logs = []
    for raw_log in raw_logs:
        timestamp_sec = raw_log['timestamp'] / 1000
        time_str = datetime.fromtimestamp(timestamp_sec).strftime('%Y-%m-%d %H:%M:%S')
        log = '%s %s' % (time_str, raw_log['message'])
        logs.append(log)

    return logs




import logging
__all__ = ['AppContext']

_logger = logging.getLogger(__name__)

class AppContext(object):

    def __init__(self, platform):
        self.platform = platform

    def dump_page_source(self, page=None):
        raise NotImplementedError()

    def take_screenshot_as_png(self, page=None):
        raise NotImplementedError()

    def log_screenshot(self, msg, *args, **kwargs):
        if hasattr(self, '_log_screenshot_delegate'):
            self._log_screenshot_delegate(msg, *args, **kwargs)
            return

        msg = msg % args
        level = kwargs['level'] if 'level' in kwargs else logging.INFO
        _logger.log(level, msg)

    def log_page_source(self, msg, *args, **kwargs):
        if hasattr(self, '_log_page_source_delegate'):
            self._log_page_source_delegate(msg, *args, **kwargs)
            return

        msg = msg % args
        level = kwargs['level'] if 'level' in kwargs else logging.INFO
        _logger.log(level, msg)

    def get_initial_logs(self):
        raise NotImplementedError()

    def get_new_logs(self):
        raise NotImplementedError()

    def quit(self):
        raise NotImplementedError()




import logging
from pyuia import PageObject
from selenium.common.exceptions import NoSuchElementException

__all__ = ['SeleniumPageObject']
_logger = logging.getLogger(__name__)

class SeleniumPageObject(PageObject):

    def __init__(self, context):
        PageObject.__init__(self, context, NoSuchElementException)

    @property
    def _driver(self):
        return self._context.driver

    def _is_displayed(self, element):
        return element.is_displayed()




import logging, os.path as path, time

__all__ = ['in_context', 'get_current_test_case', 'log_screenshot', 'log_text', 'is_test_failed']
_log = logging.getLogger(__name__)

try:
    from robot.running.context import EXECUTION_CONTEXTS as contexts
    in_context = True if contexts.current else False
except ImportError:
    in_context = False

try:
    from robot.libraries.BuiltIn import BuiltIn
    _builtin = BuiltIn()
except ImportError:
    _builtin = None

def is_test_failed():
    status = _builtin.get_variable_value('${TEST_STATUS}')
    if status == 'FAIL': return True
    assert status in ['PASS', None], status
    return False

def get_current_test_case():
    """Return a 2-tuple which uniquely identifying the current test case.

       Returns: 2-tuple (suite_source, test_name), where suite_source is an
                absolute path to the suite file or directory.
    """
    return (_builtin.get_variable_value('${SUITE_SOURCE}'),
            _builtin.get_variable_value('${TEST_NAME}'))

def _robot_logger_of_level(level):
    # standard levels: DEBUG, INFO, WARNING, ERROR, CRITICAL
    # robot levels: TRACE, DEBUG, INFO, WARN
    from robot.api import logger as robot_logger
    if level == logging.DEBUG:
        return robot_logger.debug
    elif level == logging.INFO:
        return robot_logger.info
    elif level in (logging.WARNING, logging.ERROR, logging.CRITICAL):
        return robot_logger.warn
    else:
        assert False, level

def log_screenshot(png, msg='SCREENSHOT', prefix='screenshot_', level=logging.INFO):
    filename = '%s%s.png' % (prefix, int(time.time() * 1000))
    pathname = path.join(_get_log_dir(), filename)
    with open(pathname, 'wb') as f:
	f.write(png)
    html = '<a href="%s" target="_blank"><img src="%s" width="200"></a>' % (filename, filename)

    msg = '%s<br/>%s' % (msg, html) # TODO: HTML encode msg
    _robot_logger_of_level(level)(msg, html=True)

def log_text(text, msg='TEXT', prefix='text', suffix='.txt', level=logging.INFO):
    from robot.api import logger as robot_logger

    filename = '%s%s%s' % (prefix, int(time.time() * 1000), suffix)
    pathname = path.join(_get_log_dir(), filename)
    with open(pathname, 'wb') as f:
	f.write(text.encode('utf-8'))
    html = '<a href="%s" target="_blank">%s</a>' % (filename, filename)

    msg = '%s: %s' % (msg, html) # TODO: HTML encode msg
    _robot_logger_of_level(level)(msg, html=True)

def _get_log_dir(): # TODO: one-time calculation
    variables = _builtin.get_variables()
    outdir = variables['${OUTPUT_DIR}']
    log = variables['${LOGFILE}'] # relative to the output dir
    logdir = path.dirname(log) if log != 'NONE' else '.'
    return path.abspath(path.join(outdir, logdir))




import logging, inspect
from robot.utils import ConnectionCache
from pyuia import get_page_object
from util import is_test_failed, log_screenshot, log_text, in_context as in_robot_context

__all__ = ['BaseAppLibrary', 'BasePageLibrary']
_logger = logging.getLogger(__name__)

class BaseAppLibrary(object):

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'

    def __init__(self):
        self._cache = ConnectionCache()

    def open_app(self, device_id, alias=None):
        _logger.info('Open app; device ID = [%s], alias = [%s])', device_id, alias)

        # init context and install delegates
        context = self._init_context(device_id)
        context._log_screenshot_delegate = self._log_screenshot_delegate
        context._log_page_source_delegate = self._log_page_source_delegate
        context.logs_all = [] # accumulate logs of each step

        log_text('\n'.join(context.get_initial_logs()), 'APP LOGS (Initial)', 'app_logs_initial_', '.log')
        return self._cache.register(RFConnectionCache(context), alias)

    def _init_context(self):
        raise NotImplementedError()

    def _log_screenshot_delegate(self, msg, *args, **kwargs):
        msg = msg % args
        level = kwargs['level'] if 'level' in kwargs else logging.INFO
        page = kwargs['page'] if 'page' in kwargs else None

        if page: msg += ' (%s)' % page.__class__.__name__
        log_screenshot(self._current_context.take_screenshot_as_png(), msg, level=level)

    def _log_page_source_delegate(self, msg, *args, **kwargs):
        msg = msg % args
        level = kwargs['level'] if 'level' in kwargs else logging.INFO
        page = kwargs['page'] if 'page' in kwargs else None

        if page: msg += ' (%s)' % page.__class__.__name__
        source, ext = self._current_context.dump_page_source()
        log_text(source, msg, prefix='page_source_', suffix='.%s' % ext, level=level)

    def close_app(self):
        self._cache.current.close()

    def close_all_apps(self):
        self._cache.close_all()

    def switch_device(self, alias_or_index):
        self._cache.switch(alias_or_index)

    @property
    def _current_context(self):
        return self._cache.current._context

class RFConnectionCache(object):

    def __init__(self, context):
        self._context = context

    def close(self):
        # all statements suppress possible errors, or other sessions won't be closed.
        self._capture_state()
        self._context.quit()

    def _capture_state(self):
        failed = None
        context = self._context
        try:
            failed = is_test_failed()
            msg = 'APP LOGS (about to quit, failed = %s)' % failed

            context.logs_all.extend(context.get_new_logs() )
            log_text('\n'.join(context.logs_all), msg, 'app_logs_all_', '.log')
            if not failed: return

            context.log_page_source('PAGE SOURCE (Failed)')
            context.log_screenshot('SCREENSHOT (Failed)')
        except Exception as e:
            _logger.warning('Fail to capture state. (failed = %s)', failed, exc_info=True)

def _state_capturing_decorator(method):
    def decorator(*args, **kwargs):
        name = method.__name__
        accessor = any(name.startswith(prefix) for prefix in ['is_', 'get_', 'should_be_'])
        self = args[0]

        self._capture_state()
        try:
            result = method(*args, **kwargs)
        except Exception as err:
            self._capture_state(after=True, err=err)
            raise

        if not accessor: # efficiency
            self._capture_state(after=True)
        return result

    return decorator

class _StateCapturing(type):

      def __new__(cls, clsname, bases, attrs):
          for name, obj in attrs.items():
              if not (inspect.isroutine(obj) and not name.startswith('_')): continue
              attrs[name] = _state_capturing_decorator(obj)
          return type.__new__(cls, clsname, bases, attrs)

class BasePageLibrary(object):

    if in_robot_context:
        __metaclass__ = _StateCapturing

    def __init__(self, app_library_name):
        self._app_library_name = app_library_name

    @property
    def _current_context(self):
        from robot.libraries.BuiltIn import BuiltIn
        return BuiltIn().get_library_instance(self._app_library_name)._current_context

    @property
    def _page_object(self):
        context = self._current_context
        platform = context.platform

        cls_attr = '_page_class_%s' % platform.lower()
        page_class = getattr(self, cls_attr)
        return get_page_object(page_class, context)

    def _capture_state(self, after=False, err=None):
        context = self._current_context
        try:
            # dump page source first, because it may fail to take a sceenshot (on iOS, timed out)
            if after:
                logs_step = context.get_new_logs()
                context.logs_all.extend(logs_step)
                log_text('\n'.join(logs_step), 'APP LOGS (Step)', 'app_logs_step_', '.log')

            if err: context.log_page_source('PAGE SOURCE')
            context.log_screenshot('AFTER' if after else 'BEFORE')
        except:
            _logger.warning('Fail to capture state.', exc_info=True)




from pyuia import AppContext

__all__ = ['SeleniumContext']

class SeleniumContext(AppContext):

    def __init__(self, driver):
        self.driver = driver
        platform = driver.desired_capabilities['platformName']
        AppContext.__init__(self, platform)

    def dump_page_source(self):
        return (self.driver.page_source, 'html')

    def take_screenshot_as_png(self):
        return self.driver.get_screenshot_as_png()

    def get_new_logs(self):
        raise NotImplementedError()

    def get_initial_logs(self):
        return self.get_new_logs()

    def quit(self):
        self.driver.quit()




import time
from ..selenium import SeleniumPageObject

__all__ = ['AppiumPageObject']

class AppiumPageObject(SeleniumPageObject):

    def _press_menu(self):
        self._driver.keyevent(82)
        time.sleep(1) # or then menu may not in the subsequent screenshot.

    def _press_back(self):
        self._driver.keyevent(4)




from setuptools import setup

setup(
  name = 'pyuia',
  license = 'MIT',
  packages = [
      'pyuia',
      'pyuia.robot',
      'pyuia.selenium',
      'pyuia.appium',
  ],
  version = '0.1.1',
  description = 'PyUIA is a library aiming to facilitate the implementation of UI test automation with Python.',
  author = 'KKBOX SQA Team',
  author_email = 'imsardine@gmail.com',
  url = 'https://github.com/imsardine/pyuia',
  keywords = [
      'python',
      'ui test automation',
      'keyword-driven testing',
      'page object pattern',
      'appium',
      'robot framework',
      'mobile', 'android', 'ios',
  ],
  classifiers = [
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Operating System :: OS Independent',
      'Programming Language :: Python',
      'Topic :: Software Development :: Quality Assurance',
      'Topic :: Software Development :: Testing',
  ],
)



from ..selenium import SeleniumContext
from .util import get_logs

__all__ = ['AppiumContext']

class AppiumContext(SeleniumContext):

    def dump_page_source(self):
        return (self.driver.page_source, 'xml')

    def get_new_logs(self):
        log_type = 'syslog' if self.platform == 'iOS' else 'logcat'
        return get_logs(self.driver, log_type)




import time, logging
from .exceptions import TimeoutError, ElementNotFoundError

__all__ = ['PageObject', 'get_page_object']
_logger = logging.getLogger(__name__)

_page_singletons = {} # cache

def get_page_object(page_class, context):
    fqcn = '%s.%s' % (page_class.__module__, page_class.__name__)
    _logger.debug('Get page object; FQCN = %s', fqcn)
    if fqcn in _page_singletons:
        cache = _page_singletons[fqcn]
        if cache._context is context:
            _logger.debug('Cached, and the context remains unchanged.')
            return cache
        else:
            _logger.debug('Cached, but the context is invalid.')

    page = page_class(context)
    _page_singletons[fqcn] = page
    return page

def _is_iterable(obj):
    try:
       iter(obj)
       return True
    except TypeError:
       return False

def _to_iterable(obj):
    return obj if _is_iterable(obj) else (obj,)

_WAIT_INTERVAL = 1

_WAIT_TIMEOUT = 60

_WARN_TIMEOUT = 10

_NOT_FOUND_EXCEPTIONS = (ElementNotFoundError,)

class PageObject(object):

    def __init__(self, context, not_found_exceptions):
        self._context = context

        exceptions = list(_NOT_FOUND_EXCEPTIONS)
        if not_found_exceptions:
            try:
                exceptions.extend(not_found_exceptions)
            except TypeError: # not iterable
                exceptions.append(not_found_exceptions)
        self._not_found_exceptions = tuple(exceptions)

    def _page_object(self, page_class, wait_for_loaded=True):
        """Instantiate a page object."""
        page = get_page_object(page_class, self._context)
        if wait_for_loaded: page.wait_for_page_loaded()
        return page

    def wait_for_page_loaded(self, timeout_warn=_WARN_TIMEOUT, timeout=_WAIT_TIMEOUT):
        start_time = time.time()
        timeout_warn = start_time + timeout_warn
        timeout = start_time + timeout

        while True:
            try:
                self.assert_on_this_page()
                break
            except Exception:
                # catch all exceptions, so locators that are involved in page
                # assertion can raise any exceptions (in addition to
                # AssertionError) to indicate some violation of assumptions.
                if time.time() > timeout_warn:
                    self._log_screenshot(
                        'Wait for page loaded. Time elapsed = [%s]s.',
                        time.time() - start_time,
                        level=logging.WARN)

                time.sleep(_WAIT_INTERVAL)
                if time.time() > timeout: raise
        if hasattr(self, '_wait_for_idle'):
            self._log_screenshot('Already on certain page then wait for idle.')
            self._wait_for_idle()
        return self

    def assert_on_this_page(self): pass

    def _log_screenshot(self, msg, *args, **kwargs):
        kwargs['page'] = self
        self._context.log_screenshot(msg, *args, **kwargs)

    def _log_page_source(self, msg, *args, **kwargs):
        kwargs['page'] = self
        self._context.log_page_source(msg, *args, **kwargs)

    def _assert_present(self, locators, handlers=None, check_visibility=False):
        single_loc = not _is_iterable(locators)
        locators = _to_iterable(locators)
        elements = []

        for locator in locators:
            try:
                element = locator()
            except self._not_found_exceptions as e:
                _logger.debug(
                    'Assert ALL present. The locator (%s) did not resolve to '
                    'an element.', locator, exc_info=True)
                element = None
            if not element:
                _consult_handlers(handlers, self._not_found_exceptions)
                assert False, locator # None or empty sequence

            if check_visibility and not self._is_displayed(element):
                _consult_handlers(handlers, self._not_found_exceptions)
                assert False, locator
            elements.append(element)

        return elements[0] if single_loc else elements

    def _assert_visible(self, locators, handlers=None):
        return self._assert_present(locators, handlers, check_visibility=True)

    def _assert_any_present(self, locators, handlers=None, check_visibility=False):
        locators = _to_iterable(locators)

        for locator in locators:
            try:
                element = locator()
            except self._not_found_exceptions as e:
                _logger.debug(
                    'Assert ANY present. The locator (%s) did not resolve to '
                    'an element.', locator, exc_info=True)
                element = None
            if not element: continue # None or empty sequence

            if check_visibility and not self._is_displayed(element): continue
            return element

        _consult_handlers(handlers, self._not_found_exceptions)
        assert False, locators

    def _assert_any_visible(self, locators, handlers=None):
        return self._assert_any_present(locators, handlers, check_visibility=True)

    def _wait_present(self, locators, timeout_warn=_WARN_TIMEOUT, handlers=None,
                      timeout=_WAIT_TIMEOUT, check_visibility=False):
        start_time = time.time()
        timeout_warn = start_time + timeout_warn
        timeout = start_time + timeout
        single_loc = not _is_iterable(locators)
        locators = _to_iterable(locators)

        while True:
            elements = []
            for locator in locators:
                try:
                    element = locator()
                except self._not_found_exceptions as e:
                    _logger.debug(
                        'Wait ALL present. The locator (%s) did not resolve to '
                        'an element.', locator, exc_info=True)
                    element = None
                if not element: break # None or empty sequence

                if check_visibility and not self._is_displayed(element): break
                elements.append(element)

            if len(elements) == len(locators):
                return elements[0] if single_loc else elements

            if time.time() > timeout_warn:
                self._log_screenshot(
                    'Wait ALL elements to be present. locators = %s, '
                    'check_visibility = [%s], time elapsed = [%s]s.',
                    locators, check_visibility, time.time() - start_time,
                    level=logging.WARN)
            _consult_handlers(handlers, self._not_found_exceptions)

            time.sleep(_WAIT_INTERVAL)
            if time.time() > timeout:
                raise TimeoutError(
                    'Wait ALL elements to be present. locators = %s, '
                    'check_visibility = [%s], time elapsed = [%s]s.' %
                    (locators, check_visibility, time.time() - start_time))

    def _wait_visible(self, locators, timeout_warn=_WARN_TIMEOUT, handlers=None, timeout=_WAIT_TIMEOUT):
        return self._wait_present(locators, timeout_warn, handlers, timeout, check_visibility=True)

    def _wait_any_present(self, locators, timeout_warn=_WARN_TIMEOUT, handlers=None,
                          timeout=_WAIT_TIMEOUT, check_visibility=False):
        start_time = time.time()
        timeout_warn = start_time + timeout_warn
        timeout = start_time + timeout
        locators = _to_iterable(locators)

        while True:
            for locator in locators:
                try:
                    element = locator()
                except self._not_found_exceptions as e:
                    _logger.debug(
                        'Wait ANY present. The locator (%s) did not resolve to '
                        'an element.', locator, exc_info=True)
                    element = None
                if not element: continue # None or empty sequence

                if check_visibility and not self._is_displayed(element): continue
                return element

            if time.time() > timeout_warn:
                self._log_screenshot(
                    'Wait ANY present. locators = %s, time elapsed = [%s]s.',
                    locators, time.time() - start_time, level=logging.WARN)
            _consult_handlers(handlers, self._not_found_exceptions)

            time.sleep(_WAIT_INTERVAL)
            if time.time() > timeout:
                raise TimeoutError(
                    'Wait ANY elements to be present. locators = %s, '
                    'check_visibility = [%s], time elapsed = [%s]s.' %
                    (locators, check_visibility, time.time() - start_time))

    def _wait_any_visible(self, locators, timeout_warn=_WARN_TIMEOUT, handlers=None,
                          timeout=_WAIT_TIMEOUT):
        return self._wait_any_present(locators, timeout_warn, handlers, timeout, check_visibility=True)

    def _wait_absent(self, locators, timeout_warn=_WARN_TIMEOUT, minwait=3,
                     handlers=None, timeout=_WAIT_TIMEOUT, check_visibility_only=False):
        start_time = time.time()
        timeout_appear = start_time + minwait
        timeout_warn = start_time + timeout_warn
        timeout = start_time + timeout
        locators = _to_iterable(locators)

        while True:
            # to avoid the situation that elements are absent simply because
            # other elements such as error dialogs are displayed.
            _consult_handlers(handlers, self._not_found_exceptions)
            any_invalid = False
            for locator in locators:
                try:
                    element = locator()
                except self._not_found_exceptions as e:
                    _logger.debug(
                        'Wait ALL absent. The locator (%s) did not resolve to '
                        'an element.', locator, exc_info=True)
                    element = None
                if not element: continue

                if not check_visibility_only or self._is_displayed(element):
                    any_invalid = True
                    break

            # wait for at least 'minwait' seconds to make sure target
            # element(s) won't appear at this time.
            if not any_invalid and time.time() > timeout_appear: return
            if time.time() > timeout_warn:
                self._log_screenshot(
                    'Wait ALL elements to be absent. locators = %s, '
                    'check_visibility_only = [%s], time elapsed = [%s]s.',
                    locators, check_visibility_only, time.time() - start_time, 
                    level=logging.WARN)

            time.sleep(_WAIT_INTERVAL)
            if time.time() > timeout:
                raise TimeoutError(
                    'Wait ALL elements to be absent. locators = %s, '
                    'check_visibility_only = [%s], time elapsed = [%s]s.' %
                    (locators, check_visibility_only, time.time() - start_time))

    def _wait_invisible(self, locators, timeout_warn=_WARN_TIMEOUT, minwait=3,
                        handlers=None, timeout=_WAIT_TIMEOUT):
        self._wait_absent(locators, timeout_warn, minwait, handlers, timeout, check_visibility_only=True)

    def _handle_conditional_views(self, handlers, duration=5):
        timeout = time.time() + duration
        while True:
            _consult_handlers(handlers, self._not_found_exceptions)
            if time.time() > timeout: break

def _consult_handlers(handlers, not_found_exceptions):
    if handlers is None: return

    for locator, handler in handlers.items():
        try:
            element = locator()
        except not_found_exceptions as e:
            _logger.debug(
                'Consult handlers. The locator (%s) did not resolve to an element.',
                locator, exc_info=True)
            element = None
        if not element: continue

        handler(element)

